#ifndef IDC_STATIC
#define IDC_STATIC (-1)
#endif

#define IDD_ACCMGRUI                            135
#define IDD_OPT_SIPSIMPLE                       137
#define IDD_OPT_SIPSIMPLE_ABOUT                 139
#define IDD_SEARCHUI                            142
#define IDC_USESSO                              1000
#define IDC_CONNTYPE                            1001
#define IDC_AUTHTYPE                            1002
#define IDC_ABOUTSIPE                           1003
#define IDC_IPPROGEXE                           1005
#define IDC_IPLOCALFOUND                        1006
#define IDC_PASSWORD                            1020
#define IDC_LOGIN                               1021
#define IDC_HANDLE                              1022
#define IDC_PUBLICIP                            1024
#define IDC_SSO                                 1030
#define IDC_MSO                                 1031
#define IDC_IPLOCAL                             1040
#define IDC_IPMANUAL                            1041
#define IDC_IPPROG                              1042
#define IDC_SEARCH_FN                           1044
#define IDC_SEARCH_LN                           1045
#define IDC_SEARCH_COMPANY                      1046
#define IDC_SEARCH_COUNTRY                      1047
